package CsBot3::Matrix;

use JSON qw(encode_json decode_json);
#use JSON::MaybeXS qw(encode_json decode_json);
use HTTP::Request;
use HTTP::Headers;
use LWP::UserAgent;
use URI;

our $user = "";
our $tnx_id = 0;
our $server;
our $access_token = "";
our %protocols_list = (
  'http' => 'http://',
  'https' => 'https://'
); 
our $protocol = 'https';

our %routes = (
  login => '/_matrix/client/r0/login',
  join_room => '/_matrix/client/r0/join',
  sync => '/_matrix/client/r0/sync',
  room_send_message => '/_matrix/client/r0/rooms'
);

our $request_headers = ['Content-Type' => 'application/json; charset=UTF-8'];

our $lwp = LWP::UserAgent->new();
$lwp->agent('CsMatrix');

our @rooms = ();

our @event_types = ("m.room.message");
our %events_subscriptions = ();

sub new {
  my ($server) = @_;

  $CsBot3::Matrix::server = $server;
}

sub login {
  my ($credentials) = @_; 

  $CsBot3::Matrix::user = $credentials->{user};

  my $request_url = get_request_url('login');

  my $body = encode_json($credentials);
  
  my $login_request = HTTP::Request->new(
    'POST',
    $request_url,
    $CsBot3::Matrix::request_headers,
    $body
  ) or undef;
  
  my $login_response = $CsBot3::Matrix::lwp->request($login_request) or undef;

  if(!defined $login_response) {
    return 0;
  }

  $CsBot3::Matrix::access_token = decode_json($login_response->{_content})->{access_token};

  push(@$CsBot3::Matrix::request_headers, 'Authorization', "Bearer $CsBot3::Matrix::access_token"); 

  print "[OK] - Login as $CsBot3::Matrix::user\n";
  return 1;
}

sub join_room {
  my ($room) = @_; 


  my $request_url = get_request_url('join_room');
  $request_url .= "/$room";
  $request_url .= "?access_token=$CsBot3::Matrix::access_token";

  my $room_join_request = HTTP::Request->new(
    'POST',
    $request_url,
    $CsBot3::Matrix::request_headers,
    encode_json({})
  ) or undef;
  my $room_join_response = $CsBot3::Matrix::lwp->request($room_join_request) or undef;

  if(!defined $room_join_response) {
    return 0;
  }

  push(@CsBot3::Matrix::rooms, $room);
  
  print "[OK] - Join room $room\n";
  return 1;
}

sub read_events {
  my ($channel) = @_; 

  my $request_url = get_request_url('sync');

  my $initial_sync_request = HTTP::Request->new(
    'GET',
    $request_url,
    $CsBot3::Matrix::request_headers,
    encode_json({})
  ) or undef;
  my $initial_sync_response = $CsBot3::Matrix::lwp->request($initial_sync_request) or undef;
  
  if(!defined $initial_sync_response) {
    return 0;
  }
  
  my $json = JSON->new->utf8->allow_nonref->allow_blessed->convert_blessed->pretty(1);
  my $content = decode_json($initial_sync_response->{_content})->{next_batch};

  long_poll_event(next_batch);  
  
  return 1;
}

sub get_request_url {
  my ($request_type) = @_;

  my $proto = $protocols_list{$CsBot3::Matrix::protocol};
  my $route = $CsBot3::Matrix::routes{$request_type};

  if(defined $routes{$request_type}) {
    return "$proto$server$route";
  }

  return undef;
}

sub long_poll_event {
  my ($since) = @_;
  my $timeout = 300;
  my $request_url;
  
  while (1) {
    print "[SYNC] - Events\n";
    
    $request_url = get_request_url('sync');
    $request_url .= "?since=$since&timeout=$timeout";

    $CsBot3::Matrix::lwp->timeout($timeout);

    my $initial_sync_request = HTTP::Request->new(
      'GET',
      $request_url,
      $CsBot3::Matrix::request_headers,
      encode_json({})
    ) or undef;
    
    my $initial_sync_response = $CsBot3::Matrix::lwp->request($initial_sync_request) or undef;

    if ($initial_sync_response->is_success && $initial_sync_response->{_content}) {
      my $json_response = decode_json($initial_sync_response->{_content});
      my $next_batch = $json_response->{next_batch};

      foreach my $room (@CsBot3::Matrix::rooms) {
        if(defined $json_response->{rooms}->{join}->{$room}->{timeline}->{events}) {
          $events = $json_response->{rooms}->{join}->{$room}->{timeline}->{events};

          foreach my $event (@$events) {
            foreach my $subscripted_event_name (keys %CsBot3::Matrix::events_subscriptions) {
              if(
                $event->{type} eq $subscripted_event_name &&
                time() - $event->{origin_server_ts} <= 1000 &&
                defined $CsBot3::Matrix::user &&
                $event->{sender} ne $CsBot3::Matrix::user
              ) {
                $event->{room} = $room;
                $CsBot3::Matrix::events_subscriptions{$subscripted_event_name}->($event);
              }
            }
          }
        }
      }
      $since = $next_batch;
    }
  }

  return 1;
}

sub send_message {
  my ($event) = @_; 


  my $request_url = get_request_url('room_send_message');
  $request_url .= "/$event->{room}/send/m.room.message/$CsBot3::Matrix::tnx_id";
  
  $CsBot3::Matrix::tnx_id += 1;

  my $body = {
    "msgtype" => "m.text",
    "body" => "Message received."
  };

  my $room_send_message_request = HTTP::Request->new(
    'PUT',
    $request_url,
    $CsBot3::Matrix::request_headers,
    encode_json($body)
  ) or undef;

  my $room_send_message_response = $CsBot3::Matrix::lwp->request($room_send_message_request) or undef;

  if(!defined $room_send_message_response) {
    return 0;
  }
  
  return 1;
}

sub subscribe {
  my ($event_type, $callback) = @_;

  foreach my $allowed_event (@CsBot3::Matrix::event_types) {
    if($event_type eq $allowed_event) {
      $CsBot3::Matrix::events_subscriptions{$event_type} = $callback;

      return 1;
    }
  }

  return 0;
}

1;